package org.kropp.util;

import java.util.*;

/**
 * B+ Tree leaf node.
 * 
 * @author Victor Kropp
 */
class BTreeLeafNode<K extends Comparable<K>, V> extends AbstractBTreeNode<K, V>
		implements Iterable<V> {
	/**
	 * 
	 * @param order
	 *            Tree order.
	 */
	public BTreeLeafNode(int order, AbstractBTreeNode<K, V> parent) {
		this(order, parent, null, null);
	}

	/**
	 * 
	 * @param order
	 *            Tree order
	 * @param previous
	 *            Previous leaf. <tt>null</tt> if none.
	 * @param next
	 *            Next leaf. <tt>null</tt> if none.
	 */
	public BTreeLeafNode(int order, AbstractBTreeNode<K, V> parent,
			BTreeLeafNode<K, V> previous, BTreeLeafNode<K, V> next) {
		super(order, parent);
		myEntries = new ArrayList<V>(order);
		myPrevious = previous;
		myNext = next;
	}

	/**
	 * Adds key/value pair to the leaf. Automatically splits when needed.
	 * 
	 * @param key
	 *            Key.
	 * @param value
	 *            Value.
	 */
	@Override
	public void put(K key, V value) {
		int index = indexOf(key);
		if (index == -1) {
			/*
			 * This is the first entry.
			 */
			myKeys.add(0, key);
			myEntries.add(0, value);
		} else {
			/*
			 * Adding new entry to the appropriate place.
			 */
			int rel = myKeys.get(index).compareTo(key);
			if (rel == 0) {
				myEntries.set(index, value);
			} else {
				/*
				 * if ((rel == 1) && (index > 0) ) { index--; } else
				 */
				if ((rel == -1) && (index < myOrder)) {
					index++;
				}
				myKeys.add(index, key);
				myEntries.add(index, value);
			}
		}
		((BTreeNode<K, V>) myParent).updateKeyFor(this, myKeys.get(myKeys
				.size() - 1));

		if (myKeys.size() > myOrder) {
			/*
			 * Trying to re-distribute to a sibling node.
			 */
			if (myPrevious != null) {
				if (myPrevious.myParent == myParent) {
					if (myPrevious.myKeys.size() < myOrder) {
						K newKey = myKeys.remove(0);
						myPrevious.myKeys.add(newKey);
						myPrevious.myEntries.add(myEntries.remove(0));
						((BTreeNode<K, V>) myParent).updateKeyFor(myPrevious,
								newKey);
						return;
					}
				}
			}
			if (myNext != null) {
				if (myNext.myParent == myParent) {
					if (myNext.myKeys.size() < myOrder) {
						myNext.myKeys.add(0, myKeys.remove(myKeys.size() - 1));
						myNext.myEntries.add(0, myEntries.remove(myEntries
								.size() - 1));
						((BTreeNode<K, V>) myParent).updateKeyFor(this, myKeys
								.get(myKeys.size() - 1));
						return;
					}
				}
			}

			/*
			 * If can not re-distribute, split.
			 */
			BTreeLeafNode<K, V> newLeaf = new BTreeLeafNode<K, V>(myOrder,
					myParent, this, myNext);
			myNext = newLeaf;
			for (int i = myOrder; i >= myOrder / 2; i--) {
				newLeaf.myKeys.add(0, myKeys.remove(i));
				newLeaf.myEntries.add(0, myEntries.remove(i));
			}
			((BTreeNode<K, V>) myParent).updateKeyFor(this, myKeys.get(myKeys
					.size() - 1));

			myParent.addNode(newLeaf.myKeys.get(newLeaf.myKeys.size() - 1),
					newLeaf);
		}

		((BTreeNode<K, V>) myParent).updateKeyFor(this, myKeys.get(myKeys
				.size() - 1));
	}

	/**
	 * Get value by key
	 * 
	 * @param key
	 * @return Value for given <tt>key</tt>
	 * @throws NoSuchElementException
	 *             If there is no entry with given <tt>key</tt>.
	 */
	@Override
	public V getValue(K key) {
		int found = indexOf(key);
		if (myKeys.get(found).equals(key)) {
			return myEntries.get(found);
		} else {
			throw new NoSuchElementException();
		}
	}

	/**
	 * Previous leaf node.
	 * 
	 * @return Previous leaf node.
	 */
	public BTreeLeafNode<K, V> getPrevious() {
		return myPrevious;
	}

	/**
	 * Next leaf node.
	 * 
	 * @return Next leaf node.
	 */
	public BTreeLeafNode<K, V> getNext() {
		return myNext;
	}

	/**
	 * Unsupported function.
	 */
	@Override
	void addNode(K key, AbstractBTreeNode<K, V> node) {
		throw new RuntimeException("addNode is not supported by BTreeLeafNode");
	}

	/**
	 * Finds greatest key in the node.
	 * 
	 * @return Greatest key in the node, <tt>null</tt> if empty.
	 */
	@Override
	K getMaxKey() {
		if (myEntries.size() == 0) {
			return null;
		} else {
			return myKeys.get(myKeys.size() - 1);
		}
	}

	public Iterator<V> iterator() {
		return myEntries.iterator();
	}

	private ArrayList<V> myEntries;

	private BTreeLeafNode<K, V> myPrevious;

	private BTreeLeafNode<K, V> myNext;
}
